package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.Sequence.Position;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** an efficient representation of a collection of trees (Tomita's shared packed parse forest) */
public abstract class Forest<T> /*extends PrintableTree<Forest.MyBody<T>>*/ implements Visitable<Forest.Body<T>>, IntegerMappable, GraphViz.ToGraphViz {

    private static int master_idx = 0;
    private final int idx = master_idx++;
    public int toInt() { return idx; }

    /** assume that this forest contains exactly one tree and return it; otherwise throw an exception */
    public final Tree<T> expand1() throws Ambiguous, ParseFailed {
        try {
            Iterator<Tree<T>> it = expand(true).iterator();
            if (!it.hasNext()) throw new ParseFailed();
            return it.next();
        } catch (InnerAmbiguous ia) { throw new Ambiguous(ia.f); }
    }

    /** expand this forest into a set of trees */
    public HashSet<Tree<T>> expand(boolean toss) {
        final HashSetTreeConsumer<T> ret = new HashSetTreeConsumer<T>();
        visit(new TreeMaker2<T>(toss, ret), null, null);
        if (toss && ret.size() > 1) throw new InnerAmbiguous(this);
        return ret;
    }

    private static class InnerAmbiguous extends RuntimeException {
        public final Forest<?> f;
        public InnerAmbiguous(Forest<?> f) { this.f = f; }
    }

    public static interface TreeConsumer<T> {
        public void addTree(Tree<T> t);
    }
    public static class HashSetTreeConsumer<T> extends HashSet<Tree<T>> implements TreeConsumer<T> {
        public void addTree(Tree<T> t) { super.add(t); }
    }

    static        <T> Forest<T> singleton(Input.Location loc, Position p) {
        return create(loc, null, new Forest[] { }, false, true, p); }
    static        <T> Forest<T> singleton(Input.Location loc, Forest<T> body, Position p) {
        //return create(loc, null, new Forest[] { body },  false, true, p);
        return body;
    }
    static        <T> Forest<T> leaf(Input.Location loc, T tag, Position p) { return create(loc, tag, null, false, false, p); }
    public static <T> Forest<T> create(Input.Location loc, T tag, Forest<T>[] tokens, boolean unwrap, boolean singleton, Position p) {
        return new MyBody<T>(loc, tag, tokens, unwrap, singleton, p);
    }
    // Body //////////////////////////////////////////////////////////////////////////////

    protected static interface Body<T> extends GraphViz.ToGraphViz {
        void expand(int i, TreeMaker<T> h);
    }
    public abstract void edges(GraphViz.Node n);
    protected static class MyBody<T> extends Forest<T> implements Body<T> /* extends PrintableTree<Forest<T>> implements */ {

        public boolean isTransparent() { return false; }
        public boolean isHidden() { return false; }
        public GraphViz.Node toGraphViz(GraphViz gv) {
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Node n = gv.createNode(this);
            n.label = headToString()==null?"":headToString();
            n.directed = true;
            edges(n);
            return n;
        }
        public void edges(GraphViz.Node n) {
            for(int i=0; i<tokens.length; i++) {
                if (i==tokens.length-1 && unwrap) {
                    tokens[i].edges(n);
                } else {
                    n.edge(tokens[i]);
                }
            }
        }

        public <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c) {
            ivbc.invoke(this, b, c);
        }

        private final Input.Location    location;
        private final T                 tag;
        private final Forest<T>[]       tokens;
        private final boolean           unwrap;
        private final boolean           singleton;
        private final Sequence.Position reduction;

        private MyBody(Input.Location loc, T tag, Forest<T>[] tokens, boolean unwrap, boolean singleton, Position reduction) {
            this.location = loc;
            this.tag = tag;
            this.tokens = tokens==null ? emptyForestArray : new Forest[tokens.length];
            if (tokens != null) System.arraycopy(tokens, 0, this.tokens, 0, tokens.length);
            if (tokens != null) for(int i=0; i<tokens.length; i++) if (tokens[i]==null) throw new Error(i+"");
            this.unwrap = unwrap;
            this.singleton = singleton;
            this.reduction = reduction;
        }

        public void expand(final int i, final TreeMaker<T> h) {
            if (singleton) {
                tokens[0].visit(h, null, i);
                return;
            }
            if (i==0) h.start(tag, location);

            if (i==tokens.length) {
                h.finish(tag, location);

            } else if (unwrap && i==tokens.length-1) {
                if (tokens[i] != null)
                    tokens[i].visit(h, null, 0);

            } else {
                tokens[i].visit(new TreeMaker<T>(h.toss) {
                    public void start(T head, Input.Location loc) { }
                    public void addTree(Tree<T> t) { toks.add(t); }
                    public void finish(T head, Input.Location loc) {
                        int old = h.toks.size();
                        h.addTree(new Tree<T>(loc, head, toks.toArray(tree_hint)));
                        expand(i+1, h);
                        while(h.toks.size() > old) h.toks.remove(h.toks.size()-1);
                    }
                }, null, null);
            }
        }

        protected String  headToString()         { return tag==null?null:tag.toString(); }
        protected String  headToJava()           { return null; }
        protected String  left()                 { return "{"; }
        protected String  right()                { return "}"; }
        protected boolean ignoreSingleton()      { return false; }
    }


    // Ref //////////////////////////////////////////////////////////////////////////////

    /**
     *  This class represents a partially complete collection of
     *  forests to be viewed as a forest at some later date; once
     *  viewed, it becomes immutable
     */
    static class Ref<T> extends Forest<T> {
        private FastSet<Forest<T>> hp = new FastSet<Forest<T>>();
        public Ref() { }
        public int toInt() {
            if (hp.size()==1) return hp.iterator().next().toInt();
            return super.toInt();
        }
        public void merge(Forest p) { if (p!=this) hp.add(p, true); }

        public boolean isTransparent() { return hp.size()==1; }
        public boolean isHidden() { return hp.size()==0; }
        public void edges(GraphViz.Node n) { for(Forest f : hp) f.edges(n); }
        public GraphViz.Node toGraphViz(GraphViz gv) {
            if (hp.size()==0) return null;
            if (hp.size()==1) return hp.iterator().next().toGraphViz(gv);
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Node n = gv.createNode(this);
            n.label = "?";
            n.color = "red";
            for(Forest f : hp) n.edge(f);
            return n;
        }

        public <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c) {
            if (hp==null) return;
            for(Forest<T> f : hp)
                f.visit(ivbc, b, c);
        }
        public Forest resolve() { return this; }
    }

    public abstract <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c);
    private static class TreeMaker2<T> extends TreeMaker<T> {
        private TreeConsumer<T> tc;
        public TreeMaker2(boolean toss, TreeConsumer<T> tc) { super(toss); this.tc = tc; }
        public void finish(T head, Input.Location loc) { tc.addTree(new Tree<T>(loc, head, toks.toArray(tree_hint)));; }
        public void start(T head, Input.Location loc) { }
        public void addTree(Tree<T> t) { toks.add(t); }
    }
    private static abstract class TreeMaker<T> implements Invokable<Forest.Body<T>,Boolean,Integer>, TreeConsumer<T> {
        public ArrayList<Tree<T>> toks = new ArrayList<Tree<T>>();
        private boolean toss;
        protected T head;
        public TreeMaker(boolean toss) { this.toss = toss; }
        public abstract void start(T head, Input.Location loc);
        public abstract void finish(T head, Input.Location loc);
        public abstract void addTree(Tree<T> t);
        public void invoke(Forest.Body<T> bod, Boolean o, Integer i) {
            if (i==null) {
                ArrayList<Tree<T>> toks = this.toks;
                this.toks = new ArrayList<Tree<T>>();
                bod.expand(0, this);
                this.toks = toks;
            } else {
                bod.expand(i, this);
            }
        }
    }

    // Statics //////////////////////////////////////////////////////////////////////////////

    private static Tree[] tree_hint = new Tree[0];
    private static final Forest[] emptyForestArray = new Forest[0];

    protected String  headToString()    { return null; }
    protected String  headToJava()      { return null; }
    protected String  left()            { return "<?"; }
    protected String  right()           { return "?>"; }
    protected boolean ignoreSingleton() { return true; }
}
