package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** juxtaposition; zero or more adjacent Elements; can specify a rewriting */
public abstract class Sequence extends Element implements Iterable<Element> {

    // Static Constructors //////////////////////////////////////////////////////////////////////////////

    abstract Sequence _clone();
    Sequence dup() {
        Sequence ret = _clone();
        for(Sequence s : needs) { ret.needs.add(s); s.needed.add(ret); }
        for(Sequence s : hates) { ret.hates.add(s); s.hated.add(ret); }
        ret.noFollow = noFollow;
        return ret;
    }

    /** the empty sequence (matches the empty string) */
    public static final Sequence empty = new Sequence.Constant.Empty();

    /** after matching the sequence, do not add anything to the output tree */
    public static Sequence drop(Element[] e, boolean lame) { return new Constant.Drop(e, lame); }

    /** after matching the sequence, insert a constant into the output tree */
    public static Sequence constant(Element[] e, Object o) { return new Constant(e, o); }

    /** after matching the sequence, place the result of the <tt>idx</tt>th match in the output tree */
    public static Sequence singleton(Element[] e, int idx) { return new Singleton(e, idx); }

    /**
     *  after matching the sequence, create the specified output tree
     *  @param tag   the tag for the output tree
     *  @param e     the elements to match
     *  @param drops only elements of <tt>e</tt> whose corresponding <tt>boolean</tt> in <tt>drops</tt> is <i>false</i> will be included in the output tree
     **/
    public static Sequence rewritingSequence(Object tag, Element[] e, boolean[] drops) { return new RewritingSequence(tag, e, drops); }

    ////////////////////////////////////////////////////////////////////////////////

    public Element noFollow = null;
    public final Topology noFollow() { return noFollow==null ? null : Atom.toAtom(noFollow); }

    Topology toAtom() {
        if (elements.length!=1) throw new RuntimeException("cannot invoke toAtom() on a Sequence with " + elements.length + " elements: " + this);
        return Atom.toAtom(elements[0]);
    }

    public Sequence and(Sequence s) { Sequence ret = dup(); ret.needs.add(s); s.needed.add(ret); return ret; }
    public Sequence not(Sequence s) { Sequence ret = dup(); ret.hates.add(s); s.hated.add(ret); return ret; }

    protected final Element[] elements;

    final HashSet<Sequence> needed = new HashSet<Sequence>();
    final HashSet<Sequence> hated = new HashSet<Sequence>();
    final HashSet<Sequence> needs = new HashSet<Sequence>();
    final HashSet<Sequence> hates = new HashSet<Sequence>();
    public boolean           lame  = false;

    final Position          firstp;
    Position firstp() { return firstp; }

    public Iterator<Element> iterator()    { return new ArrayIterator<Element>(elements); }
    protected Sequence(Element[] elements) {
        this.elements = elements;
        this.firstp = new Position(0);
    }

    // DO NOT MESS WITH THE FOLLOWING LINE!!!
    private Forest.Ref epsilonForm = null;
    Forest epsilonForm() {
        if (epsilonForm!=null) return epsilonForm;
        epsilonForm = new Forest.Ref();
        epsilonForm.merge(firstp().rewrite(null, false));
        return epsilonForm;
    }

    protected abstract <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args, Position p);


    // Position //////////////////////////////////////////////////////////////////////////////

    /** the imaginary position before or after an element of a sequence; corresponds to an "LR item" */
    public class Position implements IntegerMappable {

        private Forest zero = null;
        public Forest zero() {
            if (zero != null) return zero;
            if (pos > 0) throw new Error();
            return zero = rewrite(null);
        }


                final int      pos;
        private final Position next;
                final Forest[] holder;
        
        private Position(int pos) {
            this.pos      = pos;
            this.next     = pos==elements.length ? null : new Position(pos+1);
            this.holder   = new Forest[elements.length];
        }

        boolean isFirst() { return pos==0; }

        /** the element immediately after this Position, or null if this is the last Position */
        public Element  element() { return pos>=elements.length ? null : elements[pos]; }

        /** the element which produces the sequence to which this Position belongs */
        public Sequence owner() { return Sequence.this; }

        /** the next Position (the Position after <tt>this.element()</tt>) */
        public Position next() { return next; }

        /** true iff this Position is the last one in the sequence */
        public boolean isLast() { return next()==null; }

        // Position /////////////////////////////////////////////////////////////////////////////////

        final <T> Forest<T> rewrite(Input.Location loc) { return rewrite(loc, true); }
        private final <T> Forest<T> rewrite(Input.Location loc, boolean epsilonCheck) {
            if (epsilonCheck && this==firstp()) return epsilonForm();
            for(int i=0; i<pos; i++) if (holder[i]==null) throw new Error("realbad " + i);
            for(int i=pos; i<elements.length; i++) {
                if (holder[i]==null) holder[i] = elements[i].epsilonForm();
                if (holder[i]==null) throw new Error("bad " + i);
            }
            Forest<T> ret = Sequence.this.postReduce(loc, holder, this);
            for(int k=0; k<pos; k++) holder[k] = null; // to help GC
            return ret;
        }

        public String   toString() {
            StringBuffer ret = new StringBuffer();
            ret.append("<{");
            for(Position p = Sequence.this.firstp(); p != null; p = p.next()) {
                ret.append(' ');
                if (p==this) ret.append(" | ");
                if (p.element()!=null) ret.append(p.element());
                else                   ret.append(' ');
            }
            ret.append("}>");
            return ret.toString();
        }
        private final int idx = master_position_idx++;
        public int toInt() { return idx; }
    }
    private static int master_position_idx = 0;

    // toString //////////////////////////////////////////////////////////////////////////////

    public String toString() { return toString(new StringBuffer(), false).toString(); }
    StringBuffer toString(StringBuffer sb) { return toString(sb, true); }
    StringBuffer toString(StringBuffer sb, boolean spacing) {
        for(int i=0; i<elements.length; i++) {
            sb.append(elements[i].toString());
            sb.append(' ');
        }
        return sb;
    }


    // Specialized Subclasses //////////////////////////////////////////////////////////////////////////////

    static class Constant extends Sequence {
        private final Object result;
        public Constant(Element[] e, Object result) { super(e); this.result = result; }
        Sequence _clone() { return new Constant(elements, result); }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args, Position p) {
            return (Forest<T>)Forest.leaf(loc, result, p);
        }
        static class Drop extends Constant {
            Sequence _clone() { return new Drop(elements, lame); }
            public Drop(Element[] e, boolean lame) {
                super(e, null);
                this.lame = lame;
            }
        }
        static class Empty extends Sequence.Constant.Drop {
            Sequence _clone() { return new Empty(); }
            public Empty() { super(new Element[] { }, false); } }
    }

    static class Singleton extends Sequence {
        private final int idx;
        public Singleton(Element e) { this(new Element[] { e }, 0); }
        public Singleton(Element[] e, int idx) { super(e); this.idx = idx; }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args, Position p) { return (Forest<T>)Forest.singleton(loc, args[idx], p); }
        Sequence _clone() { return new Singleton(elements,idx); }
    }

    public static class Unwrap extends Sequence {
        private boolean[] drops;
        public Unwrap(Element[] e)                  { super(e); this.drops = null; }
        public Unwrap(Element[] e, boolean[] drops) { super(e); this.drops = drops; }
        Sequence _clone() { return new Unwrap(elements, drops); }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args, Position p) {
            for(int i=0; i<args.length; i++) if (args[i]==null) throw new Error();
            if (drops==null) return Forest.create(loc, null, args, true, false, p);
            int count = 0;
            for(int i=0; i<drops.length; i++) if (!drops[i]) count++;
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, null, args2, true, false, p);
        }
    }

    static class RewritingSequence extends Sequence {
        /*private*/public final Object tag;
        private final boolean[] drops;
        private int count = 0;
        Sequence _clone() { return new RewritingSequence(tag, elements, drops); }
        public RewritingSequence(Object tag, Element[] e) { this(tag, e, null); }
        public RewritingSequence(Object tag, Element[] e, boolean[] drops) {
            super(e);
            this.tag = tag;
            this.drops = drops == null ? new boolean[e.length] : drops;
            for(int i=0; i<this.drops.length; i++) if (!this.drops[i]) count++;
        }
        public <T> Forest<T> postReduce(Input.Location loc, Forest<T>[] args, Position p) {
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            //System.out.println("reduce \""+tag+"\"");
            return Forest.create(loc, (T)tag, args2, false, false, p);
        }
        public StringBuffer toString(StringBuffer sb, boolean spacing) {
            int len = sb.length();
            super.toString(sb, spacing);
            len = sb.length()-len;
            if (spacing) for(int i=0; i<50-len; i++) sb.append(' ');
            sb.append(" => ");
            sb.append(tag);
            return sb;
        }
    }
}
